==> Signal8-4.3.tar.gz: Signal8-4.3/Signal8/Signal8.py <==
import numpy as np
import matplotlib.path as mpath

from functools import partial
from .utils.scenario import BaseScenario
from .utils.simple_env import SimpleEnv, make_env
from .utils.core import Agent, Goal, Obstacle, World
from .utils.problems import get_problem_list, get_problem_instance

from gymnasium.utils import EzPickle


class raw_env(SimpleEnv, EzPickle):
    def __init__(
        self, 
        num_agents=1, 
        num_large_obstacles=4, 
        num_small_obstacles=10, 
        render_mode=None
        ):
        
        if num_agents > 2:
            raise ValueError("Signal8 currently can only support up to 2 agents.")
        
        if num_large_obstacles > 16:
            raise ValueError("Signal8 has a maximum of 10 large obstacles.")
        
        scenario = Scenario()
        world = scenario.make_world(num_agents, num_large_obstacles, num_small_obstacles)
        
        super().__init__(
            scenario=scenario, 
            world=world, 
            render_mode=render_mode,
            max_cycles=500, 
        )
        
env = make_env(raw_env)

class Scenario(BaseScenario):
    def make_world(self, num_agents, num_large_obstacles, num_small_obstacles):
        world = World()
        world.problem_list = get_problem_list()

        for i in range(num_agents):
            agent = Agent()
            agent.name = f"agent_{i}"
            agent.collide = True
            agent.color = np.array([1, 0.95, 0.8])
            world.agents.append(agent)
        
            goal = Goal()
            goal.name = f"goal_{i}"
            goal.collide = False
            goal.color = np.array([0.835, 0.90, 0.831])
            world.goals.append(goal)
        
        # Large obstacles can only be observed by aerial agent
        for i in range(num_large_obstacles):
            obstacle = Obstacle(size=0.1)
            obstacle.name = f"obs_{i}"
            obstacle.color = np.array([0.97, 0.801, 0.8])
            world.large_obstacles.append(obstacle)
        
        # Small obstacles can only be observed by ground agent(s)
        for i in range(num_small_obstacles):
            obstacle = Obstacle(size=0.01)
            obstacle.name = f"obs_{i}"
            obstacle.color = np.array([0.97, 0.801, 0.8])
            world.small_obstacles.append(obstacle)    
        
        world.buffer_dist = world.agents[0].size + world.large_obstacles[0].size
        return world
    
    # Get constraints on entities given problem instance name
    def _set_problem_instance(self, world, instance_name):
        instance_constr = get_problem_instance(instance_name)
        world.problem_instance = instance_name
        world.instance_constr = instance_constr
    
    # Generate valid points according to some condition
    def _generate_position(self, np_random, condition, circle=False):
        while True:
            if circle:
                theta = np_random.uniform(0, 2*np.pi)
                r = 0.5 * np.sqrt(np_random.uniform(0, 1))
                point = np.array([r * np.cos(theta), r * np.sin(theta)])
            else:
                point = np_random.uniform(-1, +1, 2)
            if condition(point):
                break
        return point

    # Check if point is outside of triangular obstacle regions
    def _outside_triangle(self, point, paths, epsilon):
        enlarged_paths = []
        for path in paths:
            centroid = np.mean(path.vertices[:-1], axis=0)
            enlarged_vertices = path.vertices + epsilon * (path.vertices - centroid)
            enlarged_paths.append(mpath.Path(enlarged_vertices))
        return not any(path.contains_points(point[None, :]) for path in enlarged_paths)
    
    # Check if point is outside of circular obstacle regions
    def _outside_circles(self, point, centers_radii, epsilon):
        return not any(np.linalg.norm(point - center) <= radius + epsilon for center, radius in centers_radii)
    
    # Check if point is outside of rectangular obstacle regions
    def _outside_rectangle(self, point, x_constraints, y_constraints, epsilon):
        within_constraints = any(
            (low_x - epsilon <= point[0] <= high_x + epsilon) and (low_y - epsilon <= point[1] <= high_y + epsilon)
            for (low_x, high_x), (low_y, high_y) in zip(x_constraints, y_constraints)
        )
        return not within_constraints

    # Reset agents and goals to their initial positions
    def _reset_agents_and_goals(self, world, np_random, paths):
        epsilon = world.buffer_dist
        
        if world.problem_instance not in ['einstein_tile', 'corners', 'right_arrows', 'circle', 'solar_system']:
            # Used for problem instances composed of rectangles
            x_constraints = [constr[0] for constr in world.instance_constr]
            y_constraints = [constr[1] for constr in world.instance_constr]

        for i, agent in enumerate(world.agents):
            agent.goal = world.goals[i]

            agent.state.p_vel = np.zeros(world.dim_p)
            agent.goal.state.p_vel = np.zeros(world.dim_p)

            if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
                condition = partial(
                    self._outside_triangle, 
                    paths=paths, 
                    epsilon=epsilon
                    )
            elif world.problem_instance in ['circle', 'solar_system']:
                condition = partial(
                    self._outside_circles,
                    centers_radii=world.instance_constr,
                    epsilon=epsilon
                    )
            else:
                condition = partial(
                    self._outside_rectangle, 
                    x_constraints=x_constraints, 
                    y_constraints=y_constraints,
                    epsilon=epsilon, 
                    )

            agent.state.p_pos = self._generate_position(np_random, condition)
            agent.goal.state.p_pos = self._generate_position(np_random, condition)
    
    # Reset all large obstacles to a position that does not intersect with the agents and is within its shape
    def _reset_large_obstacles(self, world, np_random, paths):
        def inside_shape_condition(point):
            return any(path.contains_points(point[None, :]) for path in paths)    
        
        occupied = set()
        num_shapes = len(world.instance_constr)
        circle = True if world.problem_instance in ['circle', 'solar_system'] else False    
        
        for i, large_obstacle in enumerate(world.large_obstacles):            
            large_obstacle.state.p_vel = np.zeros(world.dim_p)
            
            while True:
                pos = self._generate_position(np_random, inside_shape_condition, circle=circle)
                shape_idx = next((i for i, path in enumerate(paths) if path.contains_points(pos[None, :])), None)
                
                if i % num_shapes == 0:
                    occupied.clear()
                
                if shape_idx not in occupied:
                    large_obstacle.state.p_pos = pos
                    occupied.add(shape_idx)
                    break
    
    def _reset_small_obstacles(self, world, np_random, paths):
        epsilon = world.small_obstacles[0].size + world.large_obstacles[0].size

        agent_positions = [agent.state.p_pos for agent in world.agents]
        goal_positions = [goal.state.p_pos for goal in world.goals]
        large_obstacle_positions = [obstacle.state.p_pos for obstacle in world.large_obstacles]

        def safe_position(point):
            within_other_positions = any(np.linalg.norm(point - pos) <= epsilon for pos in agent_positions + goal_positions + large_obstacle_positions)

            if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
                within_obstacle_constraints = any(path.contains_points(point[None, :]) for path in paths)
            elif world.problem_instance in ['circle', 'solar_system']:
                within_obstacle_constraints = any(
                    np.linalg.norm(point - center) <= radius + epsilon 
                    for center, radius in world.instance_constr
                )
            else:
                x_constraints = [constr[0] for constr in world.instance_constr]
                y_constraints = [constr[1] for constr in world.instance_constr]
                outside_rectangle = self._outside_rectangle(point, x_constraints, y_constraints, epsilon)
                within_obstacle_constraints = not outside_rectangle

            return not (within_obstacle_constraints or within_other_positions)

        for small_obstacle in world.small_obstacles:
            small_obstacle.state.p_vel = np.zeros(world.dim_p)
            small_obstacle.state.p_pos = self._generate_position(np_random, safe_position)

    def reset_world(self, world, np_random, problem_instance):
        def make_circle_points(center, radius, num_points=100):
            t = np.linspace(0, 2*np.pi, num_points)
            x = center[0] + radius * np.cos(t)
            y = center[1] + radius * np.sin(t)
            points = np.column_stack([x, y])
            return points
        
        def make_rectangle_points(bounds):
            min_x, max_x = bounds[0]
            min_y, max_y = bounds[1]
            return [
                (min_x, min_y),
                (max_x, min_y),
                (max_x, max_y),
                (min_x, max_y),
            ]

        self._set_problem_instance(world, problem_instance)
        
        if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
            paths = [mpath.Path(np.array(triangle)) for triangle in world.instance_constr]
        elif world.problem_instance in ['circle', 'solar_system']:
            paths = [mpath.Path(make_circle_points(center, radius)) for center, radius in world.instance_constr]
        else:
            paths = [mpath.Path(make_rectangle_points(bounds)) for bounds in world.instance_constr]

        self._reset_agents_and_goals(world, np_random, paths)
        self._reset_large_obstacles(world, np_random, paths)
        self._reset_small_obstacles(world, np_random, paths)
    
    # Ground agents can only observe the positions of other agents, goals, and small obstacles
    def observation(self, agent, world):
        agent_pos = agent.state.p_pos
        
        num_agents = len(world.agents)
        num_small_obstacles = len(world.small_obstacles)
        max_observable_dist = agent.max_observable_dist
        
        observed_agents = [np.full_like(agent_pos, max_observable_dist) for _ in range(num_agents - 1)]
        observed_goal = np.full_like(agent_pos, max_observable_dist)
        observed_obstacles = [np.full_like(agent_pos, max_observable_dist) for _ in range(num_small_obstacles)]
        
        idx = 0
        for other_agent in world.agents:
            if agent.name == other_agent.name:
                continue
            else:
                other_agent_pos = other_agent.state.p_pos
                relative_pos = other_agent_pos - agent_pos
                dist = np.linalg.norm(relative_pos)
                if dist <= max_observable_dist:
                    observed_agents[idx] = relative_pos
                idx += 1
        
        for i, small_obstacle in enumerate(world.small_obstacles):
            obs_pos = small_obstacle.state.p_pos
            relative_pos = obs_pos - agent_pos
            dist = np.linalg.norm(relative_pos)
            if dist <= max_observable_dist:
                observed_obstacles[i] = relative_pos
                
        goal_pos = agent.goal.state.p_pos
        relative_goal_pos = goal_pos - agent_pos
        goal_dist = np.linalg.norm(relative_goal_pos)
        if goal_dist <= max_observable_dist:
            observed_goal = relative_goal_pos
        
        return np.concatenate((agent_pos, np.concatenate(observed_obstacles, axis=0), observed_goal))
        
    # Reward given by agents to agents for reaching their respective goals
    def reward(self, agent, world):
        return 0

==> Signal8-4.6.tar.gz: Signal8-4.6/Signal8/Signal8.py <==
import numpy as np
import matplotlib.path as mpath

from functools import partial
from .utils.scenario import BaseScenario
from .utils.simple_env import SimpleEnv, make_env
from .utils.core import Agent, Goal, Obstacle, World
from .utils.problems import get_problem_list, get_problem_instance

from gymnasium.utils import EzPickle


class raw_env(SimpleEnv, EzPickle):
    def __init__(
        self, 
        num_agents=1, 
        num_large_obstacles=4, 
        num_small_obstacles=10, 
        render_mode=None
        ):
        
        if num_agents > 2:
            raise ValueError("Signal8 currently can only support up to 2 agents.")
        
        if num_large_obstacles > 16:
            raise ValueError("Signal8 has a maximum of 10 large obstacles.")
        
        scenario = Scenario()
        world = scenario.make_world(num_agents, num_large_obstacles, num_small_obstacles)
        
        super().__init__(
            scenario=scenario, 
            world=world, 
            render_mode=render_mode,
            max_cycles=500, 
        )
        
env = make_env(raw_env)

class Scenario(BaseScenario):
    def make_world(self, num_agents, num_large_obstacles, num_small_obstacles):
        world = World()
        world.problem_list = get_problem_list()

        for i in range(num_agents):
            agent = Agent()
            agent.name = f"agent_{i}"
            agent.collide = True
            agent.color = np.array([1, 0.95, 0.8])
            world.agents.append(agent)
        
            goal = Goal()
            goal.name = f"goal_{i}"
            goal.collide = False
            goal.color = np.array([0.835, 0.90, 0.831])
            world.goals.append(goal)
        
        # Large obstacles can only be observed by aerial agent
        for i in range(num_large_obstacles):
            obstacle = Obstacle(size=0.1)
            obstacle.name = f"obs_{i}"
            obstacle.color = np.array([0.97, 0.801, 0.8])
            world.large_obstacles.append(obstacle)
        
        # Small obstacles can only be observed by ground agent(s)
        for i in range(num_small_obstacles):
            obstacle = Obstacle(size=0.01)
            obstacle.name = f"obs_{i}"
            obstacle.color = np.array([0.97, 0.801, 0.8])
            world.small_obstacles.append(obstacle)    
        
        world.buffer_dist = world.agents[0].size + world.large_obstacles[0].size
        return world
    
    # Get constraints on entities given problem instance name
    def _set_problem_instance(self, world, instance_name):
        instance_constr = get_problem_instance(instance_name)
        world.problem_instance = instance_name
        world.instance_constr = instance_constr
    
    # Generate valid points according to some condition
    def _generate_position(self, np_random, condition, circle=False, bounds=(-1, 1)):
        while True:
            if circle:
                theta = np_random.uniform(0, 2*np.pi)
                r = 0.5 * np.sqrt(np_random.uniform(0, 1))
                point = np.array([r * np.cos(theta), r * np.sin(theta)])
            else:
                point = np_random.uniform(bounds[0], bounds[1], 2)  # use the bounds here
            if condition(point):
                break
        return point

    # Check if point is outside of triangular obstacle regions
    def _outside_triangle(self, point, paths, epsilon):
        enlarged_paths = []
        for path in paths:
            centroid = np.mean(path.vertices[:-1], axis=0)
            enlarged_vertices = path.vertices + epsilon * (path.vertices - centroid)
            enlarged_paths.append(mpath.Path(enlarged_vertices))
        return not any(path.contains_points(point[None, :]) for path in enlarged_paths)
    
    # Check if point is outside of circular obstacle regions
    def _outside_circles(self, point, centers_radii, epsilon):
        return not any(np.linalg.norm(point - center) <= radius + epsilon for center, radius in centers_radii)
    
    # Check if point is outside of rectangular obstacle regions
    def _outside_rectangle(self, point, x_constraints, y_constraints, epsilon):
        within_constraints = any(
            (low_x - epsilon <= point[0] <= high_x + epsilon) and (low_y - epsilon <= point[1] <= high_y + epsilon)
            for (low_x, high_x), (low_y, high_y) in zip(x_constraints, y_constraints)
        )
        return not within_constraints

    # Reset agents and goals to their initial positions
    def _reset_agents_and_goals(self, world, np_random, paths):
        epsilon = world.buffer_dist
        
        if world.problem_instance not in ['einstein_tile', 'corners', 'right_arrows', 'circle', 'solar_system']:
            # Used for problem instances composed of rectangles
            x_constraints = [constr[0] for constr in world.instance_constr]
            y_constraints = [constr[1] for constr in world.instance_constr]

        for i, agent in enumerate(world.agents):
            agent.goal = world.goals[i]

            agent.state.p_vel = np.zeros(world.dim_p)
            agent.goal.state.p_vel = np.zeros(world.dim_p)

            if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
                condition = partial(
                    self._outside_triangle, 
                    paths=paths, 
                    epsilon=epsilon
                    )
            elif world.problem_instance in ['circle', 'solar_system']:
                condition = partial(
                    self._outside_circles,
                    centers_radii=world.instance_constr,
                    epsilon=epsilon
                    )
            else:
                condition = partial(
                    self._outside_rectangle, 
                    x_constraints=x_constraints, 
                    y_constraints=y_constraints,
                    epsilon=epsilon, 
                    )

            agent.state.p_pos = self._generate_position(np_random, condition)
            agent.goal.state.p_pos = self._generate_position(np_random, condition)
    
    # Reset all large obstacles to a position that does not intersect with the agents and is within its shape                
    def _reset_large_obstacles(self, world, np_random, paths):
        def inside_shape_condition(point):
            return any(path.contains_points(point[None, :]) for path in paths)    

        occupied = set()
        num_shapes = len(world.instance_constr)
        circle = True if world.problem_instance in ['circle', 'solar_system'] else False    
        radius = world.large_obstacles[0].size  # get the radius of the obstacle

        for i, large_obstacle in enumerate(world.large_obstacles):            
            large_obstacle.state.p_vel = np.zeros(world.dim_p)

            while True:
                pos = self._generate_position(np_random, inside_shape_condition, circle=circle, bounds=(-1+radius, 1-radius))  # adjust the bounds here
                shape_idx = next((i for i, path in enumerate(paths) if path.contains_points(pos[None, :])), None)

                if i % num_shapes == 0:
                    occupied.clear()

                if shape_idx not in occupied:
                    large_obstacle.state.p_pos = pos
                    occupied.add(shape_idx)
                    break

    def _reset_small_obstacles(self, world, np_random, paths):
        epsilon = world.small_obstacles[0].size + world.large_obstacles[0].size

        agent_positions = [agent.state.p_pos for agent in world.agents]
        goal_positions = [goal.state.p_pos for goal in world.goals]
        large_obstacle_positions = [obstacle.state.p_pos for obstacle in world.large_obstacles]

        def safe_position(point):
            within_other_positions = any(np.linalg.norm(point - pos) <= epsilon for pos in agent_positions + goal_positions + large_obstacle_positions)

            if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
                within_obstacle_constraints = any(path.contains_points(point[None, :]) for path in paths)
            elif world.problem_instance in ['circle', 'solar_system']:
                within_obstacle_constraints = any(
                    np.linalg.norm(point - center) <= radius + epsilon 
                    for center, radius in world.instance_constr
                )
            else:
                x_constraints = [constr[0] for constr in world.instance_constr]
                y_constraints = [constr[1] for constr in world.instance_constr]
                outside_rectangle = self._outside_rectangle(point, x_constraints, y_constraints, epsilon)
                within_obstacle_constraints = not outside_rectangle

            return not (within_obstacle_constraints or within_other_positions)

        for small_obstacle in world.small_obstacles:
            small_obstacle.state.p_vel = np.zeros(world.dim_p)
            small_obstacle.state.p_pos = self._generate_position(np_random, safe_position)

    def reset_world(self, world, np_random, problem_instance):
        def make_circle_points(center, radius, num_points=100):
            t = np.linspace(0, 2*np.pi, num_points)
            x = center[0] + radius * np.cos(t)
            y = center[1] + radius * np.sin(t)
            points = np.column_stack([x, y])
            return points
        
        def make_rectangle_points(bounds):
            min_x, max_x = bounds[0]
            min_y, max_y = bounds[1]
            return [
                (min_x, min_y),
                (max_x, min_y),
                (max_x, max_y),
                (min_x, max_y),
            ]

        self._set_problem_instance(world, problem_instance)
        
        if world.problem_instance in ['einstein_tile', 'corners', 'right_arrows']:
            paths = [mpath.Path(np.array(triangle)) for triangle in world.instance_constr]
        elif world.problem_instance in ['circle', 'solar_system']:
            paths = [mpath.Path(make_circle_points(center, radius)) for center, radius in world.instance_constr]
        else:
            paths = [mpath.Path(make_rectangle_points(bounds)) for bounds in world.instance_constr]

        self._reset_agents_and_goals(world, np_random, paths)
        self._reset_large_obstacles(world, np_random, paths)
        self._reset_small_obstacles(world, np_random, paths)
    
    # Ground agents can only observe the positions of other agents, goals, and small obstacles
    def observation(self, agent, world):
        agent_pos = agent.state.p_pos
        
        num_agents = len(world.agents)
        num_small_obstacles = len(world.small_obstacles)
        max_observable_dist = agent.max_observable_dist
        
        observed_agents = [np.full_like(agent_pos, max_observable_dist) for _ in range(num_agents - 1)]
        observed_goal = np.full_like(agent_pos, max_observable_dist)
        observed_obstacles = [np.full_like(agent_pos, max_observable_dist) for _ in range(num_small_obstacles)]
        
        idx = 0
        for other_agent in world.agents:
            if agent.name == other_agent.name:
                continue
            else:
                other_agent_pos = other_agent.state.p_pos
                relative_pos = other_agent_pos - agent_pos
                dist = np.linalg.norm(relative_pos)
                if dist <= max_observable_dist:
                    observed_agents[idx] = relative_pos
                idx += 1
        
        for i, small_obstacle in enumerate(world.small_obstacles):
            obs_pos = small_obstacle.state.p_pos
            relative_pos = obs_pos - agent_pos
            dist = np.linalg.norm(relative_pos)
            if dist <= max_observable_dist:
                observed_obstacles[i] = relative_pos
                
        goal_pos = agent.goal.state.p_pos
        relative_goal_pos = goal_pos - agent_pos
        goal_dist = np.linalg.norm(relative_goal_pos)
        if goal_dist <= max_observable_dist:
            observed_goal = relative_goal_pos
        
        return np.concatenate((agent_pos, np.concatenate(observed_obstacles, axis=0), observed_goal))
        
    # Reward given by agents to agents for reaching their respective goals
    def reward(self, agent, world):
        return 0
